@@ -26,17 +26,24 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <signal.h>
+#if defined(__FreeBSD__)
+#include <sys/endian.h>
+#include <sys/types.h>
+#include <net/ethernet.h>
+#else
 #include <endian.h>
+#include <netinet/ether.h>
+#endif
 #include <arpa/inet.h>
 #include <netinet/in.h>
-#include <netinet/ether.h>
 #include <sys/time.h>
 #include <time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <string.h>
-#ifdef __LINUX__
+#ifdef __linux__
 #include <linux/if_ether.h>
+#include <sys/mman.h>
 #endif
 #include "md5.h"
 #include "protocol.h"
@@ -45,6 +52,8 @@
 #include "config.h"
 #include "mactelnet.h"
 #include "mndp.h"
+#include "autologin.h"
+#include "utlist.h"
 
 #define PROGRAM_NAME "MAC-Telnet"
 
@@ -54,7 +63,7 @@ static int sockfd = 0;
 static int insockfd;
 
 static unsigned int outcounter = 0;
-static unsigned int incounter = 0;
+static long incounter = -1;
 static int sessionkey = 0;
 static int running = 1;
 
@@ -69,9 +78,15 @@ static struct in_addr destip;
 static int sourceport;
 
 static int connect_timeout = CONNECT_TIMEOUT;
+static char run_mndp = 0;
+static int mndp_timeout = 0;
 
 static int is_a_tty = 1;
 static int quiet_mode = 0;
+static int batch_mode = 0;
+static int no_autologin = 0;
+
+static char autologin_path[255];
 
 static int keepalive_counter = 0;
 
@@ -80,7 +95,7 @@ static char username[255];
 static char password[255];
 static char nonpriv_username[255];
 
-struct net_interface interfaces[MAX_INTERFACES];
+struct net_interface *interfaces=NULL;
 struct net_interface *active_interface;
 
 /* Protocol data direction */
@@ -194,6 +209,11 @@ static void send_auth(char *username, char *password) {
 	int plen;
 	md5_state_t state;
 
+#if defined(__linux__) && defined(_POSIX_MEMLOCK_RANGE)
+	mlock(md5data, sizeof(md5data));
+	mlock(md5sum, sizeof(md5data));
+#endif
+
 	/* Concat string of 0 + password + encryptionkey */
 	md5data[0] = 0;
 	strncpy(md5data + 1, password, 82);
@@ -267,7 +287,7 @@ static int handle_packet(unsigned char *data, int data_len) {
 
 		/* Accept first packet, and all packets greater than incounter, and if counter has
 		wrapped around. */
-		if (incounter == 0 || pkthdr.counter > incounter || (incounter - pkthdr.counter) > 65535) {
+		if (pkthdr.counter > incounter || (incounter - pkthdr.counter) > 65535) {
 			incounter = pkthdr.counter;
 		} else {
 			/* Ignore double or old packets */
@@ -345,33 +365,30 @@ static int find_interface() {
 	struct mt_packet data;
 	struct sockaddr_in myip;
 	unsigned char emptymac[ETH_ALEN];
-	int i, testsocket;
+	int testsocket;
 	struct timeval timeout;
 	int optval = 1;
+	struct net_interface *interface;
 
 	/* TODO: reread interfaces on HUP */
-	bzero(&interfaces, sizeof(struct net_interface) * MAX_INTERFACES);
+	//bzero(&interfaces, sizeof(struct net_interface) * MAX_INTERFACES);
 
 	bzero(emptymac, ETH_ALEN);
 
-	if (net_get_interfaces(interfaces, MAX_INTERFACES) <= 0) {
+	if (net_get_interfaces(&interfaces) <= 0) {
 		fprintf(stderr, _("Error: No suitable devices found\n"));
 		exit(1);
 	}
 
-	for (i = 0; i < MAX_INTERFACES; ++i) {
-		if (!interfaces[i].in_use) {
-			break;
-		}
-
+	DL_FOREACH(interfaces, interface) {
 		/* Skip loopback interfaces */
-		if (memcmp("lo", interfaces[i].name, 2) == 0) {
+		if (memcmp("lo", interface->name, 2) == 0) {
 			continue;
 		}
 
 		/* Initialize receiving socket on the device chosen */
 		myip.sin_family = AF_INET;
-		memcpy((void *)&myip.sin_addr, interfaces[i].ipv4_addr, IPV4_ALEN);
+		memcpy((void *)&myip.sin_addr, interface->ipv4_addr, IPV4_ALEN);
 		myip.sin_port = htons(sourceport);
 
 		/* Initialize socket and bind to udp port */
@@ -388,15 +405,15 @@ static int find_interface() {
 		}
 
 		/* Ensure that we have mac-address for this interface  */
-		if (!interfaces[i].has_mac) {
+		if (!interface->has_mac) {
 			close(testsocket);
 			continue;
 		}
 
 		/* Set the global socket handle and source mac address for send_udp() */
 		send_socket = testsocket;
-		memcpy(srcmac, interfaces[i].mac_addr, ETH_ALEN);
-		active_interface = &interfaces[i];
+		memcpy(srcmac, interface->mac_addr, ETH_ALEN);
+		active_interface = interface;
 
 		/* Send a SESSIONSTART message with the current device */
 		init_packet(&data, MT_PTYPE_SESSIONSTART, srcmac, dstmac, sessionkey, 0);
@@ -425,18 +442,22 @@ int main (int argc, char **argv) {
 	int result;
 	struct mt_packet data;
 	struct sockaddr_in si_me;
+	struct autologin_profile *login_profile;
+	struct net_interface *interface, *tmp;
 	unsigned char buff[1500];
 	unsigned char print_help = 0, have_username = 0, have_password = 0;
 	unsigned char drop_priv = 0;
 	int c;
 	int optval = 1;
 
+	strncpy(autologin_path, AUTOLOGIN_PATH, 254);
+
 	setlocale(LC_ALL, "");
 	bindtextdomain("mactelnet","/usr/share/locale");
 	textdomain("mactelnet");
 
 	while (1) {
-		c = getopt(argc, argv, "lnqt:u:p:U:vh?");
+		c = getopt(argc, argv, "lnqt:u:p:U:vh?BAa:");
 
 		if (c == -1) {
 			break;
@@ -457,6 +478,9 @@ int main (int argc, char **argv) {
 
 			case 'p':
 				/* Save password */
+#if defined(__linux__) && defined(_POSIX_MEMLOCK_RANGE)
+				mlock(password, sizeof(password));
+#endif
 				strncpy(password, optarg, sizeof(password) - 1);
 				password[sizeof(password) - 1] = '\0';
 				have_password = 1;
@@ -471,6 +495,7 @@ int main (int argc, char **argv) {
 
 			case 't':
 				connect_timeout = atoi(optarg);
+				mndp_timeout = connect_timeout;
 				break;
 
 			case 'v':
@@ -483,7 +508,19 @@ int main (int argc, char **argv) {
 				break;
 
 			case 'l':
-				return mndp();
+				run_mndp = 1;
+				break;
+
+			case 'B':
+				batch_mode = 1;
+				break;
+
+			case 'A':
+				no_autologin = 1;
+				break;
+
+			case 'a':
+				strncpy(autologin_path, optarg, 254);
 				break;
 
 			case 'h':
@@ -493,9 +530,12 @@ int main (int argc, char **argv) {
 
 		}
 	}
+	if (run_mndp) {
+		return mndp(mndp_timeout, batch_mode);
+	}
 	if (argc - optind < 1 || print_help) {
 		print_version();
-		fprintf(stderr, _("Usage: %s <MAC|identity> [-h] [-n] [-t <timeout>] [-u <user>] [-p <password>] [-U <user>] | -l\n"), argv[0]);
+		fprintf(stderr, _("Usage: %s <MAC|identity> [-h] [-n] [-a <path>] [-A] [-t <timeout>] [-u <user>] [-p <password>] [-U <user>] | -l [-B] [-t <timeout>]\n"), argv[0]);
 
 		if (print_help) {
 			fprintf(stderr, _("\nParameters:\n"
@@ -503,9 +543,12 @@ int main (int argc, char **argv) {
 			"                 discover it.\n"
 			"  identity       The identity/name of your destination device. Uses\n"
 			"                 MNDP protocol to find it.\n"
-			"  -l             List/Search for routers nearby. (using MNDP)\n"
+			"  -l             List/Search for routers nearby (MNDP). You may use -t to set timeout.\n"
+			"  -B             Batch mode. Use computer readable output (CSV), for use with -l.\n"
 			"  -n             Do not use broadcast packets. Less insecure but requires\n"
 			"                 root privileges.\n"
+			"  -a <path>      Use specified path instead of the default: " AUTOLOGIN_PATH " for autologin config file.\n"
+			"  -A             Disable autologin feature.\n"
 			"  -t <timeout>   Amount of seconds to wait for a response on each interface.\n"
 			"  -u <user>      Specify username on command line.\n"
 			"  -p <password>  Specify password on command line.\n"
@@ -523,6 +566,30 @@ int main (int argc, char **argv) {
 		quiet_mode = 1;
 	}
 
+	if (!no_autologin) {
+		autologin_readfile(autologin_path);
+		login_profile = autologin_find_profile(argv[optind]);
+
+		if (!quiet_mode && login_profile != NULL && (login_profile->hasUsername || login_profile->hasPassword)) {
+			fprintf(stderr, _("Using autologin credentials from %s\n"), autologin_path);
+		}
+		if (!have_username) {
+			if (login_profile != NULL && login_profile->hasUsername) {
+				have_username = 1;
+				strncpy(username, login_profile->username, sizeof(username) - 1);
+				username[sizeof(username) - 1] = '\0';
+			}
+		}
+
+		if (!have_password) {
+			if (login_profile != NULL && login_profile->hasPassword) {
+				have_password = 1;
+				strncpy(password, login_profile->password, sizeof(password) - 1);
+				password[sizeof(password) - 1] = '\0';
+			}
+		}
+	}
+
 	/* Seed randomizer */
 	srand(time(NULL));
 
@@ -568,6 +635,7 @@ int main (int argc, char **argv) {
 	if (!have_username) {
 		if (!quiet_mode) {
 			printf(_("Login: "));
+			fflush(stdout);
 		}
 		scanf("%254s", username);
 	}
@@ -575,11 +643,14 @@ int main (int argc, char **argv) {
 	if (!have_password) {
 		char *tmp;
 		tmp = getpass(quiet_mode ? "" : _("Password: "));
+#if defined(__linux__) && defined(_POSIX_MEMLOCK_RANGE)
+		mlock(password, sizeof(password));
+#endif
 		strncpy(password, tmp, sizeof(password) - 1);
 		password[sizeof(password) - 1] = '\0';
 		/* security */
 		memset(tmp, 0, strlen(tmp));
-#ifdef __GNUC__
+#ifdef __linux__
 		free(tmp);
 #endif
 	}
@@ -592,7 +663,7 @@ int main (int argc, char **argv) {
 	inet_pton(AF_INET, (char *)"255.255.255.255", &destip);
 	memcpy(&sourceip, &(si_me.sin_addr), IPV4_ALEN);
 
-	/* Sessioon key */
+	/* Session key */
 	sessionkey = rand() % 65535;
 
 	/* stop output buffering */
@@ -692,5 +763,10 @@ int main (int argc, char **argv) {
 	close(sockfd);
 	close(insockfd);
 
+	DL_FOREACH_SAFE(interfaces, interface, tmp) {
+		DL_DELETE(interfaces, interface);
+		free(interface);
+	}
+
 	return 0;
 }
